﻿//////////////////////////////////////////////
// BufferCast.h
//
//////////////////////////////////////////////

/// Defines / Macros -------------------------

#pragma once

/// Includes ---------------------------------

// nkMemory
#include "Buffer.h"

// Standards
#include <array>
#include <vector>

/// Internals --------------------------------

namespace nkMemory
{
	template <typename T>
	struct BufferCastDataDescriptor final
	{
		T* _data ;
		unsigned long long _size ;
	} ;
}

/// Class ------------------------------------

namespace nkMemory
{
	template <typename T = unsigned char>
	class BufferCast
	{
		private :

			// Attributes
			// Base buffer
			T* _data ;
			unsigned long long _size ;

		public :

			// Constructors
			BufferCast ()
			:	_data (nullptr),
				_size (0)
			{
				// Nothing to do
			}
			
			BufferCast (unsigned long long size)
			:	_data (size ? new T [size] : nullptr),
				_size (size)
			{
				// Nothing to do
			}

			BufferCast (T* data, unsigned long long size)
			:	_data (size ? new T [size] : nullptr),
				_size (size)
			{
				if constexpr (std::is_class_v<T>)
				{
					for (unsigned long long i = 0 ; i < _size ; ++i)
						_data[i] = data[i] ;
				}
				else
				{
					if (_size)
						memcpy(_data, data, size * sizeof(T)) ;
				}
			}

			BufferCast (const BufferCast& cast)
			:	_data (cast._size ? new T [cast._size] : nullptr),
				_size (cast._size)
			{
				// Copy data around
				if constexpr (std::is_class_v<T>)
				{
					for (unsigned long long i = 0 ; i < _size ; ++i)
						_data[i] = cast._data[i] ;
				}
				else
				{
					if (_size)
						memcpy(_data, cast._data, _size * sizeof(T)) ;
				}
			}

			BufferCast (BufferCast&& cast)
			:	_data (cast._data),
				_size (cast._size)
			{
				// Reset input
				cast._data = nullptr ;
				cast._size = 0 ;
			}

			template <typename U = T>
			BufferCast (std::enable_if_t<!std::is_class_v<U>, const Buffer&> buffer)
			:	_data (buffer.getSize() ? new T [buffer.getSize() / sizeof(T)] : nullptr),
				_size (buffer.getSize() / sizeof(T))
			{
				// Copy data around
				if (_size)
					memcpy(_data, buffer.getData(), _size * sizeof(T)) ;
			}

			// Destructor
			~BufferCast ()
			{
				delete[] _data ;
				_data = nullptr ;
			}

		public :

			// Getters
			T* getData () const
			{
				return _data ;
			}

			unsigned long long getSize () const
			{
				return _size ;
			}

			bool empty () const
			{
				return _size == 0 ;
			}

			T& front ()
			{
				return _data[0] ;
			}

			const T& front () const
			{
				return _data[0] ;
			}

			T& back ()
			{
				return _data[_size - 1] ;
			}

			const T& back () const
			{
				return _data[_size - 1] ;
			}

			T* begin ()
			{
				return _data ;
			}

			const T* begin () const
			{
				return _data ;
			}

			T* end ()
			{
				return &_data[_size] ;
			}

			const T* end () const
			{
				return &_data[_size] ;
			}

		public :

			// Management
			typename
			BufferCast<T>& clear ()
			{
				delete[] _data ;

				_data = nullptr ;
				_size = 0 ;

				return *this ;
			}

			typename
			BufferCast<T>& resize (unsigned long long size)
			{
				// Safety check
				if (!size)
				{
					clear() ;
					return *this ;
				}

				// Copy into new memory spot
				T* newPtr = new T [size] ;

				unsigned long long range = (size < _size) ? size : _size ;

				// Check if we have a complex type or not
				if constexpr (std::is_class_v<T>)
				{
					// Check if we can move or not
					if constexpr (std::is_move_assignable_v<T>)
					{
						for (unsigned long long i = 0 ; i < range ; ++i)
							newPtr[i] = std::move(_data[i]) ;
					}
					else
					{
						for (unsigned long long i = 0 ; i < range ; ++i)
							newPtr[i] = _data[i] ;
					}
				}
				else
					memcpy(newPtr, _data, range * sizeof(T)) ;

				delete[] _data ;
				_data = newPtr ;
				_size = size ;

				return *this ;
			}

			T& append (const T& value)
			{
				unsigned long long size = getSize() ;
				resize(size + 1) ;

				_data[size] = value ;

				return _data[size] ;
			}

			template <typename U = T>
			typename std::enable_if_t<std::is_class_v<U>, T&>
			append (T&& value)
			{
				unsigned long long size = getSize() ;
				resize(size + 1) ;

				_data[size] = std::forward<T>(value) ;

				return _data[size] ;
			}

			typename
			BufferCast<T>& erase (unsigned long long index, unsigned long long count = 1ull)
			{
				// Do we clear or trim right side ?
				if (!count)
					return *this ;
				else if (index == 0 && count == _size)
					clear() ;
				else if (index + count >= _size)
					resize(index) ;
				else
				{
					// Allocate new memory
					T* newData = new T [_size - count] ;

					if constexpr (std::is_class_v<T>)
					{
						if constexpr (std::is_move_assignable_v<T>)
						{
							for (unsigned long long i = 0 ; i < index ; ++i)
								newData[i] = std::move(_data[i]) ;

							unsigned long long secondStart = index + count ;

							for (unsigned long long i = 0 ; i < _size - secondStart ; ++i)
								newData[index + i] = std::move(_data[secondStart + i]) ;
						}
						else
						{
							for (unsigned long long i = 0 ; i < index ; ++i)
								newData[i] = _data[i] ;

							unsigned long long secondStart = index + count ;

							for (unsigned long long i = 0 ; i < _size - secondStart ; ++i)
								newData[index + i] = _data[secondStart + i] ;
						}
					}
					else
					{
						// Copy left and right sides around erase memory
						memcpy(newData, _data, index * sizeof(T)) ;
						memcpy(newData + index, _data + index + count, (_size - index - count) * sizeof(T)) ;
					}

					// Free and update
					delete[] _data ;
					_data = newData ;
					_size -= count ;
				}

				return *this ;
			}

			template <typename U = T>
			typename std::enable_if_t<!std::is_class_v<U> || std::is_trivially_destructible_v<U>, Buffer>
			relinquishBufferOwnership ()
			{
				Buffer result = Buffer::createAndTakeMemory((unsigned char*)_data, _size * sizeof(T)) ;

				_data = nullptr ;
				_size = 0 ;

				return result ;
			}

			typename
			BufferCastDataDescriptor<T> relinquishDataOwnership ()
			{
				BufferCastDataDescriptor<T> result ;
				result._data = _data ;
				result._size = _size ;

				_data = nullptr ;
				_size = 0 ;

				return result ;
			}

		public :

			// Operators
			T& operator[] (unsigned long long index)
			{
				return _data[index] ;
			}

			const T& operator[] (unsigned long long index) const
			{
				return _data[index] ;
			}
			
			BufferCast<T>& operator= (const BufferCast<T>& other)
			{
				// Safety for self assignment
				if (_data == other._data)
					return *this ;

				clear() ;

				// If empty, stop there
				if (other.empty())
					return *this ;

				_data = new T [other._size] ;
				_size = other._size ;

				if constexpr (std::is_class_v<T>)
				{
					for (unsigned long long i = 0 ; i < _size ; ++i)
						_data[i] = other._data[i] ;
				}
				else
					memcpy(_data, other._data, _size * sizeof(T)) ;

				return *this ;
			}

			template <typename U = T>
			std::enable_if_t<!std::is_class_v<U> || std::is_move_assignable_v<U>, BufferCast<T>&>
			operator= (BufferCast<T>&& other)
			{
				// Safety for self assignment
				if (_data == other._data)
					return *this ;

				// Clear ourselves and copy data over here
				clear() ;

				// Safety check
				if (other.empty())
					return *this ;

				_data = new T [other._size] ;
				_size = other._size ;

				if constexpr (std::is_class_v<T>)
				{
					for (unsigned long long i = 0 ; i < _size ; ++i)
						_data[i] = std::move(other._data[i]) ;
				}
				else
					memcpy(_data, other._data, _size * sizeof(T)) ;

				// Clear other array
				other.clear() ;

				return *this ;
			}

		public :

			// Templated constructors
			template <std::size_t S>
			BufferCast (const std::array<T, S>& array)
			:	_data (!array.empty() ? new T [S] : nullptr),
				_size (S)
			{
				// Copy data around
				if constexpr (std::is_class_v<T>)
				{
					for (unsigned long long i = 0 ; i < _size ; ++i)
						_data[i] = array[i] ;
				}
				else
				{
					if (!array.empty())
						memcpy(_data, array.data(), _size * sizeof(T)) ;
				}
			}

			template <typename U, std::size_t S>
			BufferCast (const std::array<U, S>& array)
			:	_data (!array.empty() ? new T [S] : nullptr),
				_size (S)
			{
				// Copy around
				for (unsigned long long i = 0 ; i < S ; ++i)
					_data[i] = array[i] ;
			}

			BufferCast (const std::vector<T>& vec)
			:	_data (!vec.empty() ? new T [vec.size()] : nullptr),
				_size (vec.size())
			{
				// Copy data around
				if constexpr (std::is_class_v<T>)
				{
					for (unsigned long long i = 0 ; i < _size ; ++i)
						_data[i] = vec[i] ;
				}
				else
				{
					if (_size)
						memcpy(_data, vec.data(), _size * sizeof(T)) ;
				}
			}

			template <typename U>
			BufferCast (const std::vector<U>& vec)
			:	_data (!vec.empty() ? new T [vec.size()] : nullptr),
				_size (vec.size())
			{
				// Copy data around
				for (unsigned long long i = 0 ; i < _size ; ++i)
					_data[i] = vec[i] ;
			}
	} ;
}

/// Exports ----------------------------------

namespace nkMemory
{
	// Unsigned
	template class DLL_MEMORY_EXPORT BufferCast<unsigned char> ;
	template class DLL_MEMORY_EXPORT BufferCast<unsigned short> ;
	template class DLL_MEMORY_EXPORT BufferCast<unsigned int> ;
	template class DLL_MEMORY_EXPORT BufferCast<unsigned long long> ;

	// Signed
	template class DLL_MEMORY_EXPORT BufferCast<char> ;
	template class DLL_MEMORY_EXPORT BufferCast<short> ;
	template class DLL_MEMORY_EXPORT BufferCast<int> ;
	template class DLL_MEMORY_EXPORT BufferCast<long long> ;

	// Floating
	template class DLL_MEMORY_EXPORT BufferCast<float> ;
	template class DLL_MEMORY_EXPORT BufferCast<double> ;
}